@@ -16,7 +16,7 @@ require 5.004;
 
 use strict;
 use vars qw($VERSION);
-$VERSION = '0.1.4';
+$VERSION = '0.1';
 
 ######################################################################
 
@@ -33,7 +33,7 @@ $VERSION = '0.1.4';
 =head2 Nonstandard Modules
 
         LWP::UserAgent 0.01
-        Cache::Cache 1.02
+        IPC::SharedCache 1.3
 
 =cut
 
@@ -42,32 +42,51 @@ $VERSION = '0.1.4';
 use Switch;
 use Carp;
 use LWP::UserAgent;
-use Cache::SharedMemoryCache;
+use IPC::SharedCache;
 
 ######################################################################
 
 =head1 SYNOPSIS
 
-    use XML::ParseDTD;
-    $dtd = XML::ParseDTD->new($dtd);
-    $bool = $dtd->child_allowed($tag, $childtag);
-    $bool = $dtd->child_list_allowed($tag, @childtags);
-    $bool = $dtd->attr_allowed($tag, $attribute);
-    $bool = $dtd->attr_list_allowed($tag, @attributes);
-    $bool = $dtd->is_empty($tag);
-    $bool = $dtd->is_defined($tag);
-    $bool = $dtd->is_fixed($tag, $attribute);
-    $bool = $dtd->attr_value_allowed($tag, $attribute, $value);
-    $bool = $dtd->attr_list_value_allowed($tag, \%attribute_value);
-    @tags = $dtd->get_document_tags();
-    $regexp = $dtd->get_child_regexp($tag);
-    @attributes = $dtd->get_attributes($tag);
-    @req_attributes = $dtd->get_req_attributes($tag);
-    $value = $dtd->get_allowed_attr_values($tag, $attribute);
-    $default_value = $dtd->get_attr_def_value($tag, $attribute);
-    $dtd->clear_cache();
-    $errormessage = $dtd->errstr;
-    $errornumber = $dtd->err;
+use XML::ParseDTD;
+
+$dtd = XML::ParseDTD->new($dtd);
+
+$bool = $dtd->child_allowed($tag, $childtag);
+
+$bool = $dtd->child_list_allowed($tag, @childtags);
+
+$bool = $dtd->attr_allowed($tag, $attribute);
+
+$bool = $dtd->attr_list_allowed($tag, @attributes);
+
+$bool = $dtd->is_empty($tag);
+
+$bool = $dtd->is_defined($tag);
+
+$bool = $dtd->is_fixed($tag, $attribute);
+
+$bool = $dtd->attr_value_allowed($tag, $attribute, $value);
+
+$bool = $dtd->attr_list_value_allowed($tag, \%attribute_value);
+
+@tags = $dtd->get_document_tags();
+
+$regexp = $dtd->get_child_regexp($tag);
+
+@attributes = $dtd->get_attributes($tag);
+
+@req_attributes = $dtd->get_req_attributes($tag);
+
+$value = $dtd->get_allowed_attr_values($tag, $attribute);
+
+$default_value = $dtd->get_attr_def_value($tag, $attribute);
+
+$dtd->clear_cache();
+
+$errormessage = $dtd->errstr;
+
+$errornumber = $dtd->err;
 
 =head1 DESCRIPTION
 
@@ -79,19 +98,17 @@ child tags a certain tag might have, which tags are defined as a empty
 tag, which attributes a certain tag might have, which values are
 allowed for a certain attribute, which attributes are required, which
 attributes are fixed, which attributes have which default value
-... well i would say it tells you all except the entity definitions
-(they're on the ToDo list) that is defined in the dtd (at least all
-that i know of, but i'm not so much into that topic, so please make me
-aware if i missed something). All this information can be accessed in
-2 diffrent ways: 1. you can simply get it 2. you can pass certain data
-and the module then tells you whether thats ok or not.
+... well i would say it tells you all except the entity definitions (they're on the ToDo list) that is defined in the dtd (at
+least all that i know of, but i'm not so much into that topic, so
+please make me aware if i missed something). All this information can
+be accessed in 2 diffrent ways: 1. you can simply get it 2. you can
+pass certain data and the module then tells you whether thats ok or
+not.
 
-This package uses Cache::SharedMemoryCache to cache every parsed DTD,
-so  next time the data structure representing the dtd can be just
-taken out of memory. Thus the dtd is not refetched and not parsed
-again which saves quite some time and work. You can easily modify the
-module so that is uses Cache::FileCache if you prefer, but i think
-SharedMemory is faster.
+This package uses IPC::SharedCache to cache every parsed DTD, so 
+next time the data structure representing the dtd can be just taken out of
+memory. Thus the dtd is not refetched and not parsed again which saves
+quite some time and work.
 
 Everytime the constructor is called it first checks whether the given
 dtd is already in memory, if so it compares the I<last modified> date
@@ -135,7 +152,7 @@ file://home/moritz/xhtml1-strict.dtd
 =back
 
 The configuration hash can be used to influence the modules
-behaviour. The options known are:
+behaviour. So far only one configuration option is known:
 
 =over
 
@@ -144,58 +161,28 @@ behaviour. The options known are:
 B<checklm> - configures how often the I<Last-Modified> header should
 be checked if the http protocol is used. The Default is I<3> that
 means that averaged it is checked every third time (dtd is refetched
-and reparsed if it was modified meanwhile). Setting it to 1 will
+and reparsed if it was modified meanwhile). Setting it to 1 or 0 will
 force the module to always check the I<Last-Modified> header, setting
 it to -1 will force it to never check the header (which is recommend
 if performance is important and its more or less sure that the dtd
 will not be changed).
 
-B<memkey> - Identifier for the datastructure which is saved to and
-taken from the cache. Because this module uses the shared memory for
-caching, it is important that is identifier is really unique, else it
-would probably overwrite some data of another program. By default the
-Identifier is I<XML::ParseDTD>. I</URL of the parsed dtd> is allways
-added to the value of this option to distinguish the dtds.
-
-B<timeout> - The value of this option is simply passed to
-LWP::UserAgent as timeout value. Please see the documentation of
-LWP::UserAgent for more information. The default is
-I<30>. LWP::UserAgent is used to fetch dtds with the http protocol and
-to get their I<last modified> header to know whether they have been
-modified.
-
-B<cache_expire> - The value of this option is passed to Cache::Cache
-for setting the time when the cache will be expired and thus has to be
-rewritten. By default this is I<never>. For possible values please
-read the documentation of Cache::Cache.
-
 =back
 
-B<Note>: You shouldn't set any option to I<0> since it will not be interpreted, that means the default setting will be used instead.
-
 =cut
 
 ######################################################################
 
+my $checklm = 3;
+my $ipc_key = 'XML::ParseDTD';
+
 sub new {
-  my $checklm = 3;
-  my $memkey = 'XML::ParseDTD';
-  my $timeout = 30;
-  my $cache_expire = 'never';
   my ($class, $dtd, %conf) = @_;
-  #file:// or without protocol is the same (local filesystem)
-  $dtd =~ s/^file:\/\///;
-  my $cache = new Cache::SharedMemoryCache;
-  #attach dtd path to key
-  $memkey = ($conf{memkey}||$memkey).'/'.$dtd;
-  $cache->purge();
-  my $self = $cache->get($memkey);
-  if(!defined($self) || !_validate($dtd,$self,$conf{checklm}||$checklm,$conf{timeout}||$timeout)) {
-    $self = _load($dtd,$conf{timeout}||$timeout);
-    $cache->set($memkey,$self,$conf{cache_expire}||$cache_expire);
-  }
-  $self = bless($self, ref($class) || $class);
-  $self->{cache} = $cache;
+  $checklm = $conf{checklm} if(defined($conf{checklm}));
+  my %cache;
+  tie %cache, 'IPC::SharedCache', ipc_key => $ipc_key, load_callback => \&_load, validate_callback => \&_validate;
+  $_ = $cache{$dtd};
+  my $self = bless($_, ref($class) || $class);
   return $self;
 }
 
@@ -216,7 +203,7 @@ Returns 1 (true) or 0 (false).
 sub child_allowed {
   my($self,$tag,$child) = @_;
   $self->_set_errstr(1,$tag) and return 0 unless($self->{'Element'}->{$tag});
-  return 1 if (eval("'" . $self->{'Element'}->{$tag} . "'" . "=~ m/\($child,\)/"));
+  return 1 if (eval("'" . $self->{'Element'}->{$tag} . "'" . "=~ m/($child,)/"));
   $self->_set_errstr(4,$child,$tag);
   return 0;
 }
@@ -319,25 +306,6 @@ sub is_empty {
 
 ######################################################################
 
-=head3 is_any ($tag)
-
-Checks whether the given tag has content model I<ANY>.
-
-Returns 1 (true) or 0 (false).
-
-=cut
-
-######################################################################
-
-sub is_any {
-  my($self,$tag) = @_;
-  return 1 if($self->{'Any'}->{$tag});
-  $self->_set_errstr(11, $tag);
-  return 0;
-}
-
-######################################################################
-
 =head3 is_defined ($tag)
 
 Checks whether the given tag is defined in the dtd, that means whether
@@ -580,16 +548,15 @@ sub get_attr_def_value {
 
 =head3 clear_cache ()
 
-Clears the cache, that means that the dtd will be refetched and
-reparsed next time.
+Clears the cache, that means that all dtds will be refetched and
+reparsed.
 
 =cut
 
 ######################################################################
 
 sub clear_cache {
-  my $self = shift;
-  $self->{cache}->clear();
+  IPC::SharedCache::remove $ipc_key;
 }
 
 ######################################################################
@@ -647,20 +614,42 @@ sub _get_errstr {
         || /^8$/  && sprintf('"%s" is not a empty tag', @_)
         || /^9$/  && sprintf('Attribute "%s" for "%s" is not fixed', @_)
         || /^10$/ && sprintf('No default value defined for attribute "%s" of "%s"', @_)
-	|| /^11$/ && sprintf('Content model of tag "%s" is NOT "ANY"', @_)
 	|| '';
   }
   return $msg;
 }
 
+#sub _check_id {
+#  $_ = shift;
+#  return 0 if(m/^[A-Za-z_]{1}[A-Za-z0-9_:.-]*$/ && ! m/^(xml|XML)/);
+#  return 'must be of type ID which means it must match ^[A-Za-z]{1}[A-Za-z0-9_:.-]*$ and mustn\'t begin with xml or XML';
+#}
+
+#sub _check_idrefs {
+#  $_ = shift;
+#  return 0 if(m/^[A-Za-z_]{1}[A-Za-z0-9_:. -]*$/ && ! m/(^| )(xml|XML)/);
+#  return 'must be of type IDREFS which means it must match ^[A-Za-z]{1}[A-Za-z0-9_:. -]*$ and mustn\'t begin with xml or XML';
+#}
+
+#sub _check_cdata {
+#  return 0 ;
+#}
+
+#sub _pcdata {
+#  return 0 ;
+#}
+
+#sub _check_nmtoken {
+#  return 0 if(shift =~ m/^[A-Za-z0-9_:.-]{1}\S*$/);
+#  return 'must be of type NMTOKEN which means it must match ^[A-Za-z0-9_:.-]{1}\S*$';
+#}
 
 #this method fetches and parses the dtd
 sub _load {
-  my ($dtd,$timeout) = @_;
+  my $dtd = shift;
   my %pdtd = (
 	      'Element' => {},
 	      'Empty' => {},
-	      'Any' => {},
 	      'Attr' => {},
 	      'ReqAtt' => {},
 	      'FixAtt' => {},
@@ -668,13 +657,14 @@ sub _load {
 	     );
   my $DTD;
   if($dtd =~ m/^(?!file)([A-za-z]+):\/\//i) {
-    my $ua = LWP::UserAgent->new(timeout => $timeout);
+    my $ua = LWP::UserAgent->new(timeout => 30);
     local $_;
     $_ = $ua->get($dtd);
     $DTD = $_->content;
     $pdtd{lmod} = $_->last_modified;
   }
   else {
+    $dtd =~ s/^file:\/\///;
     open DTD, "<$dtd" or die "Cannot open file $dtd : $!\n";
     {
       local $/;
@@ -700,7 +690,7 @@ sub _load {
   #$DTD =~ s/%(\S+);/$IntEntity{$1}/gs;
   while($DTD =~ s/%(\S+);/$IntEntity{$1}/s) {}
 
-  while($DTD =~ s/<!ELEMENT\s*(\S+)\s*(?:(\([^<>]*\)(\*|\+)?)|(EMPTY)|(ANY))\s*>//s) {
+  while($DTD =~ s/<!ELEMENT\s*(\S+)\s*(?:(\([^<>]*\)(\*|\+)?)|(EMPTY))\s*>//s) {
     if(!$4) {
       $_ = $1;
       $pdtd{'Element'}->{$_} = $2;
@@ -710,14 +700,8 @@ sub _load {
       $pdtd{'Element'}->{$_} =~ s/([^a-zA-Z0-9#]{1}),/$1/gs;
     }
     else {
-      if($4 eq 'EMPTY') {
-	$pdtd{'Element'}->{$1} = 1;
-	$pdtd{'Empty'}->{$1} = 1;
-      }
-      elsif($4 eq 'ANY') {
-	$pdtd{'Element'}->{$1} = '.*';
-	$pdtd{'Any'}->{$1} = 1;
-      }
+      $pdtd{'Element'}->{$1} = 1;
+      $pdtd{'Empty'}->{$1} = 1;
     }
   }
   
@@ -727,7 +711,7 @@ sub _load {
     $pdtd{'Attr'}->{$elem} = {};
     $_ = $2;
     my ($attr,$type,$some,$default);
-    while(s/\s*(\S+)\s*((?:\([^\(\)]+\))|(?:[^\(\) \t\n]+))\s*(\S+)?\s*((?:"|')\S+(?:'|"))?\s*//s) {
+    while(s/\s*(\S+)\s*((?:\([^\(\)]+\))|(?:[^\(\) \n]+))\s*(\S+)?\s*((?:"|')\S+(?:'|"))?\s*//s) {
       ($attr,$type,$some,$default) = ($1,$2,$3,$4);
       for($type) {
 	#/^ID(REF)?$/ && do { $pdtd{'Attr'}->{$elem}->{$attr} = \&XML::ParseDTD::_check_id; last; };
@@ -770,10 +754,10 @@ sub _load {
 
 #this method proves whether the dtd is already cached and if so if it should be refetched (and reparsed)
 sub _validate {
-  my ($dtd,$rec,$checklm,$timeout) = @_;
+  my ($dtd,$rec) = @_;
   my $lmod;
-  if($dtd =~ m/^([A-za-z]+):\/\//i) {
-    $lmod = ($checklm < 0 || int(rand($checklm))) ? $rec->{lmod} : LWP::UserAgent->new(timeout => $timeout)->head($dtd)->last_modified;
+  if($dtd =~ m/^(?!file)([A-za-z]+):\/\//i) {
+    $lmod = ($checklm < 0 || int(rand($checklm))) ? $rec->{lmod} : LWP::UserAgent->new(timeout => 1)->head($dtd)->last_modified;
   }
   else {
     $lmod = (stat($dtd))[9];
@@ -787,9 +771,7 @@ __END__
 
 =head1 BUGS
 
-Send bug reports to: bug-XML-ParseDTD@rt.cpan.org (if that doesn't
-work feel free to send directly to moritz@freesources.org). Or use the
-webinterface at http://rt.cpan.org/NoAuth/Bugs.html?Dist=XML-ParseDTD.
+Send bug reports to: moritz@freesources.org
 
 Thanks!
 
